import json
import jsonschema
import re
import warnings
import yaml

import importlib.resources as pkg_resources

from copy import deepcopy
from dataengineeringutils3.s3 import read_json_from_s3, read_yaml_from_s3
from mojap_metadata.metadata import specs
from typing import Union, List, Callable, Iterator
from collections.abc import MutableMapping


_table_schema = json.load(pkg_resources.open_text(specs, "table_schema.json"))
_schema_url = "https://moj-analytical-services.github.io/metadata_schema/mojap_metadata/v1.3.0.json"  # noqa

_metadata_struct_dtype_names = ("struct",)
_metadata_struct_dtype_names_bracket = tuple(
    [x + "<" for x in _metadata_struct_dtype_names]
)
_metadata_list_dtype_names = ("list", "list_", "large_list", "array")
_metadata_list_dtype_names_bracket = tuple(
    [x + "<" for x in _metadata_list_dtype_names]
)
_metadata_complex_dtype_names = (
    _metadata_struct_dtype_names + _metadata_list_dtype_names
)
_metadata_complex_dtype_names_bracket = (
    _metadata_struct_dtype_names_bracket + _metadata_list_dtype_names_bracket
)


def _get_type_category_pattern_dict_from_schema():
    out = {}
    for type_cat in _table_schema["definitions"].keys():
        if not type_cat.endswith("types"):
            continue

        for a in _table_schema["definitions"][type_cat]["allOf"]:
            enum = a.get("properties", {}).get("type_category", {}).get("enum")
            patt = a.get("properties", {}).get("type", {}).get("pattern")

            if enum and patt:
                new_patt = patt.replace("\\\\", "\\")
                out[enum[0]] = rf"{new_patt}"
                break

    return out


def _parse_and_split(text: str, char: str) -> List[str]:
    """
    Splits a string into a list by splitting on
    any input char that is outside of parentheses.
    If `char` is inside parentheses then no split
    occurs. Also strips each str in the list.
    """
    in_parentheses = [0, 0, 0]  # [square, round, angular]

    start = -1
    for i, s in enumerate(text):
        if s == "[":
            in_parentheses[0] += 1
        elif s == "]":
            in_parentheses[0] -= 1
        elif s == "(":
            in_parentheses[1] += 1
        elif s == ")":
            in_parentheses[1] -= 1
        elif s == "<":
            in_parentheses[2] += 1
        elif s == ">":
            in_parentheses[2] -= 1

        if s == char and not any([bool(p) for p in in_parentheses]):
            yield text[start + 1 : i].strip()
            start = i

    yield text[start + 1 :].strip()


def _get_first_level(text: str) -> str:
    """Returns everything in first set of <>"""
    bracket_counter = 0
    start = -1
    end = -1
    found_first_bracket = False
    for i, c in enumerate(text):
        if c == "<":
            bracket_counter += 1
            if not found_first_bracket:
                start = i + 1
                found_first_bracket = True
        elif c == ">":
            bracket_counter -= 1
            if bracket_counter == 0:
                end = i
                break

    if start == -1 or end == -1:
        raise ValueError(f"No closed brackets found in: {text}")

    return text[start:end]


def _unpack_complex_data_type(data_type: str) -> Union[str, dict]:
    """Recursive function that jumps into complex
    data types and returns complex types as a dict.
    Non complex types are returned as a str.

    Args:
        data_type (str): Name of agnostic data type

    Returns:
        Union[str, dict]: unpacked representation of data type as dict
            for complex types. If datatype is not complex then original
            data type is returned (as str).
    """
    d = {}
    _metadata_list_dtype_names_bracket
    _metadata_struct_dtype_names_bracket
    _metadata_complex_dtype_names_bracket

    if data_type.startswith(_metadata_struct_dtype_names_bracket):
        d["struct"] = {}
        next_data_type = _get_first_level(data_type)
        for data_param in _parse_and_split(next_data_type, ","):
            k, v = data_param.split(":", 1)
            k = k.strip()
            v = v.strip()
            if v.startswith(_metadata_complex_dtype_names_bracket):
                d["struct"][k] = _unpack_complex_data_type(v)
            else:
                d["struct"][k] = v
        return d
    elif data_type.startswith(_metadata_list_dtype_names_bracket):
        k = data_type.split("<", 1)[0]
        d[k] = {}
        next_data_type = _get_first_level(data_type).strip()
        if next_data_type.startswith(_metadata_complex_dtype_names_bracket):
            d[k] = _unpack_complex_data_type(next_data_type)
        else:
            d[k] = next_data_type
        return d
    else:
        return data_type


class MetadataProperty:
    def __set_name__(self, owner, name) -> None:
        self.name = name

    def __get__(self, obj, type=None) -> object:
        return obj.__dict__["_data"].get(self.name)

    def __set__(self, obj, value) -> None:
        obj.__dict__["_data"][self.name] = value
        obj.validate()


class Metadata(MutableMapping):
    @classmethod
    def from_dict(cls, d: dict) -> object:
        """
        generates Metadata object from a dictioanary representation of metadata
        args:
            d: metadata that adheres to the metadata schema
        returns:
            Metadata object
        """
        m = cls()
        m._init_data_with_default_key_values(d)
        m.validate()
        return m

    @classmethod
    def from_json(
        cls, filename: str, encoding: str = "utf-8", *args, **kwargs
    ) -> object:
        """
        generates Metadata object from a string path to a json metadata formatted file
        args:
            filename: path to the metadata json file
        returns:
            Metadata object
        """
        if filename.startswith("s3://"):
            obj = read_json_from_s3(filename, encoding, *args, **kwargs)
        else:
            with open(filename, "r") as f:
                obj = json.load(f, **kwargs)

        return cls.from_dict(obj)

    @classmethod
    def from_yaml(
        cls, filename: str, encoding: str = "utf-8", *args, **kwargs
    ) -> object:
        """
        generates Metadata object from a string path to a yaml metadata formatted file
        args:
            filename: path to the metadata yaml file
        returns:
            Metadata object
        """
        if filename.startswith("s3://"):
            obj = read_yaml_from_s3(filename, encoding, *args, **kwargs)
        else:
            with open(filename, "r") as f:
                obj = yaml.safe_load(f, **kwargs)

        return cls.from_dict(obj)

    @classmethod
    def from_infer(cls, inp: Union[str, dict, object], **kwargs) -> object:
        """
        generates Metadata object from and infers the format of the metadata input and
        uses one of the above class methods to load. If passed a Metadata object, will
        return a copy.
        args:
            inp: input that is either a string path to a yaml or json file, a dictionary
            or a Metadata object
        returns:
            Metadata object
        """
        if isinstance(inp, str) and inp.lower().endswith("json"):
            return cls.from_json(inp, **kwargs)
        elif isinstance(inp, str) and inp.lower().endswith("yaml"):
            return cls.from_yaml(inp, **kwargs)
        elif isinstance(inp, dict):
            return cls.from_dict(inp)
        elif isinstance(inp, cls):
            return deepcopy(inp)
        else:
            raise TypeError(f"input type not recognised: {type(inp)}")

    @classmethod
    def merge(
        cls,
        old: Union[str, dict, object],
        new: Union[str, dict, object],
        mismatch: str = "priority",
        data_override: Union[dict, None] = None,
    ) -> object:
        """
        Creates a new Metadata object by merging two others.
        args:
            old: Metadata object
            new: Metadata object
            mismatch: Determines behaviour when there is a mismatch of column
                metadata, "priority" prioritises the details from the
                new Metadata object, "error" throws an error
            data: Metadata values for the merged Metadata object, overriding
                existing values
        returns:
            Metadata object
        """
        old_meta = cls.from_infer(old)
        old_meta.set_col_types_from_type_category()
        new_meta = cls.from_infer(new)
        new_meta.set_col_types_from_type_category()

        if data_override is None:
            data_override = {}
        for k in data_override:
            old_meta._data[k] = data_override[k]

        if mismatch == "error":
            parameters_to_check = [
                x
                for x in old_meta._data
                if x in new_meta._data and x != "columns" and x not in data_override
            ]
            for param in parameters_to_check:
                if old_meta._data[param] != new_meta._data[param]:
                    raise ValueError(
                        f"Merge error: values of {param} do not match,"
                        f" {old_meta._data[param]} != {new_meta._data[param]}"
                    )
            cols_to_check = set(old_meta.column_names) & set(new_meta.column_names)
            for c in cols_to_check:
                old_i = old_meta.column_names.index(c)
                new_i = new_meta.column_names.index(c)
                if old_meta.columns[old_i] != new_meta.columns[new_i]:
                    raise ValueError(
                        f"Merge error: values of column {c} do not match,"
                        f" {old_meta.columns[old_i]} != {new_meta.columns[new_i]}"
                    )

        # Update columns from new metadata
        for c in new_meta.columns:
            old_meta.update_column(c)
        # Update the rest of the data from new metadata unless specified in params
        for param in [
            x for x in new_meta._data if x != "columns" and x not in data_override
        ]:
            old_meta._data[param] = new_meta._data[param]
        old_meta.validate()
        return old_meta

    name = MetadataProperty()
    description = MetadataProperty()
    file_format = MetadataProperty()
    sensitive = MetadataProperty()
    primary_key = MetadataProperty()
    database_name = MetadataProperty()
    table_location = MetadataProperty()

    def __init__(
        self,
        name: str = "",
        description: str = "",
        file_format: str = "",
        sensitive: bool = False,
        columns: list = None,
        primary_key: list = None,
        partitions: list = None,
        force_partition_order: str = None,
    ) -> None:

        self._schema = deepcopy(_table_schema)

        self._data = {
            "$schema": _schema_url,
            "name": name,
            "description": description,
            "file_format": file_format,
            "sensitive": sensitive,
            "columns": columns if columns else [],
            "primary_key": primary_key if primary_key else [],
            "partitions": partitions if partitions else [],
        }

        self.default_type_category_lookup = {
            "null": "null",
            "integer": "int64",
            "float": "float64",
            "string": "string",
            "timestamp": "timestamp(s)",
            "binary": "binary",
            "boolean": "bool",
            "list": "list<null>",
            "struct": "struct<null>",
        }

        self.validate()
        self.force_partition_order = force_partition_order

    @property
    def force_partition_order(self):
        return self._force_partition_order

    @force_partition_order.setter
    def force_partition_order(self, order: str):
        if order is not None and order not in ["start", "end"]:
            raise ValueError(
                "force_partition_order can only be set to None, "
                f'"start" or "end". Given {order}'
            )
        else:
            self._force_partition_order = order
            self.reorder_cols_based_on_partition_order()

    def reorder_cols_based_on_partition_order(self):
        """
        Reorders columns if metadata data has columns, partitions
        and has force_partition_order set to "start" or "end".
        """
        if self.force_partition_order and self.partitions and self.columns:
            non_partitions = [
                c for c in self.columns if c["name"] not in self.partitions
            ]
            partitions = [c for c in self.columns if c["name"] in self.partitions]

            # Ensure partitions listed in column match partition order
            partitions = sorted(
                partitions, key=lambda x: self.partitions.index(x["name"])
            )

            # Set private property to avoild recursive calling
            # no validation takes place as no data is added (just reordered)
            if self.force_partition_order == "start":
                self._data["columns"] = partitions + non_partitions
            else:
                self._data["columns"] = non_partitions + partitions
        else:
            pass

    @property
    def columns(self):
        return self._data["columns"]

    @columns.setter
    def columns(self, columns: List[dict]):
        self._data["columns"] = columns
        self.validate()
        self.reorder_cols_based_on_partition_order()

    @property
    def partitions(self):
        return self._data["partitions"]

    @partitions.setter
    def partitions(self, partitions: List[str]):
        self._data["partitions"] = partitions
        self.validate()
        self.reorder_cols_based_on_partition_order()

    @property
    def column_names(self):
        return [c["name"] for c in self.columns]

    def get_column(self, name: str):
        """
        Returns a column thats name matched input.
        None is returned if no match.
        """
        c = None
        for col in self.columns:
            if col["name"] == name:
                c = col
                break
        return c

    def remove_column(self, name: str):
        if name in self.column_names:
            del self.columns[self.column_names.index(name)]
            if name in self.partitions:
                del self.partitions[self.partitions.index(name)]
        else:
            raise ValueError(f"Column: {name} not in metadata columns")

    def update_column(self, column: dict):
        """
        Adds a column to the columns property. If the column name
        alredy exists in in the metadata then that column is replaced.
        If the column name does not exist it is added to the end.

        Args:
            column (dict): A dict that has the expected properties of
                a column.
        """
        name = column["name"]
        if name in self.column_names:
            i = self.column_names.index(name)
            self.columns[i] = column
        else:
            self.columns.append(column)
        self.validate()

    def _init_data_with_default_key_values(self, data: dict):
        """
        Used to create the class from a dictionary

        Args:
            data (dict): [description]
            copy_data (bool, optional): [description]. Defaults to True.
        """
        _data = deepcopy(data)
        self._data = _data

        defaults = {
            "$schema": _schema_url,
            "name": "",
            "description": "",
            "file_format": "",
            "sensitive": False,
            "columns": [],
            "primary_key": [],
            "partitions": [],
        }

        for k, v in defaults.items():
            self._data[k] = _data.get(k, v)

    def validate(self):
        jsonschema.validate(instance=self._data, schema=self._schema)
        self._validate_list_attribute(attribute="primary_key", columns=self.primary_key)
        self._validate_list_attribute(attribute="partitions", columns=self.partitions)
        # Ensure unique column names
        self._validate_list_attribute(
            attribute="column name", columns=self.column_names
        )

    def _validate_list_attribute(self, attribute: str, columns: list) -> None:
        if not isinstance(columns, list):
            raise TypeError(f"'{attribute}' must be of type 'list'")
        if not all([isinstance(column, str) for column in columns]):
            raise TypeError(f"'{attribute}' must be a list of strings")
        if not set(columns).issubset(
            {column["name"] for column in self.columns if self.columns}
        ):
            raise ValueError(f"'All elements of '{attribute}' must be in self.columns")
        if len(columns) != len(set(columns)):
            raise ValueError(f"'All elements of '{attribute}' must be unique")

    def to_dict(self) -> dict:
        return deepcopy(self._data)

    def to_json(self, filepath: str, mode: str = "w", **kwargs) -> None:
        with open(filepath, mode) as f:
            json.dump(self.to_dict(), f, **kwargs)

    def to_yaml(self, filepath: str, mode: str = "w", **kwargs) -> None:
        with open(filepath, mode) as f:
            yaml.safe_dump(self.to_dict(), f)

    def column_names_to_lower(self, inplace: bool = False) -> Union[object, None]:
        if inplace:
            for c in self.columns:
                c["name"] = c["name"].lower()
            return self
        else:
            self_lower = deepcopy(self)
            for c in self_lower.columns:
                c["name"] = c["name"].lower()
            return self_lower

    def column_names_to_upper(self, inplace: bool = False) -> Union[object, None]:
        if inplace:
            for c in self.columns:
                c["name"] = c["name"].upper()
            return self
        else:
            self_upper = deepcopy(self)
            for c in self_upper.columns:
                c["name"] = c["name"].upper()
            return self_upper

    def unpack_complex_data_type(self, data_type: str) -> Union[str, dict]:
        """
        Takes the coltype definition as a string and parses it.
        If the data_type is complex (list_ or struct) then a dict
        representation of unpacked coltypes is returned. If the
        type is not complex then the same input `data_type` is
        returned.

        Args:
            coltype (str): Agnostic data type

        Returns:
            Union[str, dict]: unpacked representation of data type as dict
                for complex types. If data type is not complex then original
                data type is returned (as str).
        """
        return _unpack_complex_data_type(data_type)

    def set_col_type_category_from_types(self):
        """
        Sets any missing type_category based on the type attribute of the column.
        """
        mapping = _get_type_category_pattern_dict_from_schema()

        # Apply new types
        for col in self.columns:
            if "type_category" not in col and col.get("type"):
                for type_cat, pattern in mapping.items():
                    if re.match(pattern, col["type"]):
                        col["type_category"] = type_cat
                        break

    def set_col_types_from_type_category(self, type_category_lookup: Callable = None):
        """Set any missing type attribute for each column
        based on the type_category attribute.

        Args:
            type_category_lookup (Callable): A function that takes
            the col dict and returns the type based on attributes of the column
            dict. To apply a simple dictionary lookup pass the dictionary get
            method with a lambda as this arg (e.g. lambda x: d.get(x["type_category"])).
            If None applies the dictionary (self.default_type_category_lookup) is used
            to assign types based on the column's type_category attribute. Will raise a
            warning for "list" or "struct" type_category (this is due to these types
            need) to be specified by the user.
        """

        # Define standard getter for type category lookup
        if type_category_lookup is None:

            def type_category_lookup(col):
                tc = col.get("type_category")
                if tc is None:
                    err_msg = (
                        "type and type_category attributes "
                        f"are missing from the col: {name}"
                    )
                    raise KeyError(err_msg)

                if tc in ["struct", "list"]:
                    warn_msg = (
                        "For type_category of struct or list only a basic "
                        "version is inferred i.e. struct<null> or list<null>"
                    )
                    warnings.warn(warn_msg)

                return self.default_type_category_lookup.get(tc)

        # Apply new types
        for col in self.columns:
            name = col.get("name")
            if col.get("type") is None:
                new_type = type_category_lookup(col)

                if new_type is None:
                    raise ValueError(f"No type returned for col: {col}")
                col["type"] = new_type

        self.validate()

    # Metadata is a subclass of collections.abc.MutableMapping class
    # with keys as metadata column names, and values as columns
    def __getitem__(self, __key: str) -> dict:
        return self.get_column(__key)

    def __delitem__(self, __key: str) -> None:
        self.remove_column(__key)

    def __setitem__(self, __key: str, __value: dict) -> None:
        if __key != __value["name"]:
            raise ValueError(f"Column name {__value['name']} doesn't match key {__key}")
        else:
            self.update_column(__value)

    def __iter__(self) -> Iterator[str]:
        for col in self.columns:
            yield col

    def __len__(self) -> int:
        return len(self.columns)



from sqlalchemy import create_engine


def create_postgres_connection(db_settings: dict, database_name: str = "postgres"):
    """Create SQL Alchemy engine for a postgres database.

    Get connection details from file specified in config.json

    Parameters
    ----------
    db_settings (dict):
        Dictionary of database connection details.

    database_name (str):
        Specify a database to connect to within the RDS instance.
    """
    user = db_settings["user"]
    password = db_settings["password"]
    host = db_settings["host"]
    port = db_settings["port"]
    database_name = database_name
    print(f"Connecting to database {database_name}")

    # See SQLAlchemy docs for creating the db_string
    db_string = f"postgresql+psycopg2://{user}:{password}@{host}:{port}/{database_name}"
    return create_engine(db_string, echo=True)  # verbose output (sqlalchemy logging)



import warnings

from mojap_metadata.metadata.metadata import (
    Metadata,
    _unpack_complex_data_type,
    _metadata_complex_dtype_names,
)
from mojap_metadata.converters import (
    BaseConverter,
    _flatten_and_convert_complex_data_type,
)

import pyarrow as pa
from typing import Tuple, List, Any, Union, Callable


_arrow_id_to_static_metatype = {
    0: "null",
    1: "bool",
    2: "uint8",
    3: "int8",
    4: "uint16",
    5: "int16",
    6: "uint32",
    7: "int32",
    8: "uint64",
    9: "int64",
    10: "float16",
    11: "float32",
    12: "float64",
    13: "string",
    14: "binary",  # Non fixed width
    16: "date32",
    17: "date64",
    34: "large_string",
    35: "large_binary",
}


def _get_arrow_time(arrow_type: Union[pa.lib.Time32Type, pa.lib.Time64Type]):
    return f"time{arrow_type.bit_width}({arrow_type.unit})"


def _get_arrow_timestamp(arrow_type: pa.lib.TimestampType):
    return f"timestamp({arrow_type.unit})"


def _get_decimal128(arrow_type: pa.lib.Decimal128Type):
    return f"decimal128({arrow_type.precision}, {arrow_type.scale})"


def _get_fixed_width_binary(arrow_type: pa.lib.FixedSizeBinaryType):
    return f"binary({arrow_type.byte_width})"


_arrow_id_to_callable_metatype = {
    15: _get_fixed_width_binary,
    18: _get_arrow_timestamp,
    19: _get_arrow_time,
    20: _get_arrow_time,
    23: _get_decimal128,
}


def _rename_data_type_to_arrow_type(data_type: str):
    if data_type == "bool":
        return "bool_"
    elif data_type == "list":
        return "list_"
    else:
        return data_type


def _simple_arrow_type_conversion(arrow_type: pa.lib.DataType) -> str:
    if arrow_type.id in _arrow_id_to_static_metatype:
        mtype = _arrow_id_to_static_metatype[arrow_type.id]
    elif arrow_type.id in _arrow_id_to_callable_metatype:
        mtype = _arrow_id_to_callable_metatype[arrow_type.id](arrow_type)
    else:
        raise NotImplementedError(
            f"No conversion available for arrow_type: {str(arrow_type)}"
        )
    return mtype


def _is_pa_struct(thing: Any):
    return isinstance(thing, pa.lib.StructType)


def _is_pa_list(thing: Any):
    return isinstance(thing, (pa.lib.ListType, pa.lib.LargeListType))


def _convert_complex_pa_to_data_type(arrow_type: pa.lib.DataType) -> Union[dict, str]:
    """Recursive function that jumps into complex arrow types (structs and lists)
    and returns complex types as a dict. Non complex types are returned as a str.
    Names returned are those given to the agnostic data types in Metadata.

    Args:
        arrow_type (pa.lib.DataType): An arrow type

    Returns:
        Union[dict, str]: unpacked representation of data type as dict
            or str.
    """
    d = {}
    # For struct iterate through each key value binding
    # until each field name has its corresponding type returned
    if _is_pa_struct(arrow_type):
        d["struct"] = {}
        for field in arrow_type:
            d["struct"][field.name] = _convert_complex_pa_to_data_type(field.type)
        return d
    # For list get the data type for the lists content and return the list
    elif _is_pa_list(arrow_type):
        k = "list" if arrow_type.id == 25 else "large_list"
        d[k] = _convert_complex_pa_to_data_type(arrow_type.value_type)
        return d
    else:
        # If the data type is simple then return the simple conversion
        # this return is what ends the recursion
        return _simple_arrow_type_conversion(arrow_type)


def _convert_complex_data_type_to_pa(
    data_type: Union[dict, str], converter_fun: Callable
) -> Any:
    """
    Recursive function to unpack complex and basic datatypes.
    """
    if isinstance(data_type, str):
        return converter_fun(data_type)

    else:
        fields = []
        for k, v in data_type.items():
            if k in _metadata_complex_dtype_names:
                inner_data_type = _convert_complex_data_type_to_pa(v, converter_fun)
                arrow_attr = _rename_data_type_to_arrow_type(k)
                return getattr(pa, arrow_attr)(inner_data_type)
            else:
                new_v = _convert_complex_data_type_to_pa(v, converter_fun)
                fields.append((k, new_v))

        return fields


def _extract_bracket_params(meta_type: str) -> Tuple[str, List[Any]]:
    """
    Gets parameters from the string representation of the type

    Args:
        meta_type (str): The string name of the metadata type

    Returns:
        Tuple[str, List[Any]]: A tuple, first arg is a string of the type
        name only and then the second value is a list of values (if any)
        inside the brackets of the meta_type. e.g. "int64" returns ("int64", [])
        and "decimal128(1,2)" returns ("decimal128", [1, 2])
    """
    is_decimal_type = meta_type.startswith("decimal128")
    is_binary_type = meta_type.startswith("binary")

    if "(" in meta_type:
        attr_name, value_str = meta_type.split("(", 1)
        value_str = value_str.split(")")[0]
        values = value_str.split(",")
        if not any([bool(v) for v in values]):
            values = []

        # cast input to int for specific types
        if (is_decimal_type or is_binary_type) and values:
            values = [int(v.strip()) for v in values]
    else:
        attr_name = meta_type
        values = []
    return attr_name, values


class ArrowConverter(BaseConverter):
    def __init__(self):
        """
        Converts metadata objects to an Arrow Schema.
        Note that this converter has no options
        (i.e. ArrowConverter().options returns
        the BaseCoverterOptions)

        Example:
        from mojap_metadata.converters.arrow_converter import (
            ArrowConverter,
        )

        ac = ArrowConverter()
        metadata = Metadata.from_json("my-table-metadata.json")
        pyarrow_schema = ac.generate_from_meta(metadata) # get pyArrow Schema
        """
        super().__init__(None)

    def convert_col_type(self, coltype: str) -> pa.DataType:
        """Converts our metadata types to arrow data type object

        Args:
            coltype (str): str representation of our metadata column types

        Returns:
            pa.DataType: Arrow data type
        """

        data_type = _unpack_complex_data_type(coltype)

        return _convert_complex_data_type_to_pa(data_type, self.convert_basic_col_type)

    def convert_basic_col_type(self, coltype: str) -> pa.DataType:
        """
        Returns a pyarrow type from the name
        of the type in our metadata (which is based on
        arrow types)

        Args:
            coltype ([str]): str representation of our metadata column types

        Returns:
            [pa.DataType]: The equivalent type object in pyArrow
        """
        is_time_type = coltype.startswith("time")
        is_decimal_type = coltype.startswith("decimal128")
        is_binary_type = coltype.startswith("binary")
        if is_time_type or is_decimal_type or is_binary_type:
            attr_name, values = _extract_bracket_params(coltype)
        else:
            attr_name = coltype
            values = []

        # Allowing for types without underscores
        if coltype == "bool":
            attr_name = "bool_"
        elif coltype == "list":
            attr_name = "list_"
        else:
            pass

        return getattr(pa, attr_name)(*values)

    def generate_from_meta(
        self,
        metadata: Metadata,
        drop_partitions: bool = True,
    ) -> pa.Schema:
        """Generates an arrow schema from our metadata

        Args:
            metadata (Metadata): metadata object from the Metadata class
            drop_partitions (bool): Drop partitions from the outputted pyarrow schema.
              defaults to True.

        Returns:
            pa.Schema: A Schema for a pyArrow table
        """

        arrow_cols = []
        for col in metadata.columns:
            if drop_partitions and (col["name"] in metadata.partitions):
                pass
            else:
                arrow_cols.append(
                    pa.field(
                        col["name"],
                        self.convert_col_type(col["type"]),
                        nullable=col.get("nullable", True),
                    )
                )

        return pa.schema(arrow_cols)

    def generate_to_meta(
        self, arrow_schema: pa.Schema, meta_init_dict: dict = None
    ) -> Metadata:
        """Generates our metadata instance from an arrow schema

        Args:
            arrow_schema (pa.Schema): pa.Schema from an arrow table

        Returns:
            Metadata: An agnostic metadata instance
        """
        if not meta_init_dict:
            meta_init_dict = {}

        if "columns" in meta_init_dict:
            warnings.warn("columns key found in meta_init_dict will be overwritten")

        meta_init_dict["columns"] = []
        meta_init_dict["_converted_from"] = "arrow_schema"

        for field in arrow_schema:
            meta_init_dict["columns"].append(
                {"name": field.name, "type": self.reverse_convert_col_type(field.type)}
            )

        m = Metadata.from_dict(meta_init_dict)
        return m

    def reverse_convert_col_type(self, arrow_type: pa.lib.DataType) -> str:
        """Converts an arrow type to a metadata col type

        Args:
            arrow_field (pa.Field): an arrow field to convert

        Returns:
            str: str representation of Metadata col type
        """
        data_type_dict = _convert_complex_pa_to_data_type(arrow_type)
        # Types in are data_type_dict have already beenm converted
        # to our agnostic metadata type names so can flatten with
        # no converter (aka use this lambda fun)
        data_type = _flatten_and_convert_complex_data_type(
            data_type=data_type_dict, converter_fun=lambda x: x
        )
        return data_type



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mojap_metadata',
 'mojap_metadata.converters',
 'mojap_metadata.converters.arrow_converter',
 'mojap_metadata.converters.etl_manager_converter',
 'mojap_metadata.converters.glue_converter',
 'mojap_metadata.converters.glue_converter.specs',
 'mojap_metadata.converters.postgres_converter',
 'mojap_metadata.metadata',
 'mojap_metadata.metadata.specs']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1',
 'dataengineeringutils3>=1.4.0',
 'jsonschema>=3.0.0',
 'parameterized>=0.7.0,<0.8.0']

extras_require = \
{'arrow': ['pyarrow>=3.0.0'],
 'etl-manager': ['etl-manager>=7.4,<8.0'],
 'glue': ['pydbtools>=5.3.0,<6.0.0'],
 'postgres': ['psycopg2[postgres]>=2.9.2,<3.0.0',
              'testing.postgresql[postgres]>=1.3.0,<2.0.0',
              'SQLAlchemy[postgres]>=1.4.27,<2.0.0']}

setup_kwargs = {
    'name': 'mojap-metadata',
    'version': '1.11.1',
    'description': 'A python package to manage metadata',
    'long_description': '# mojap-metadata\n\nThis python package allows users to read and alter our metadata schemas (using the metadata module) as well as convert our metadata schemas to other schema definitions utilised by other tools (these are defined in the converters module and are defined as Converters).\n\n\n## Installation\n\nMake sure you are using a new version of pip (>=20.0.0)\n\n```bash\npip install git+https://github.com/moj-analytical-services/mojap-metadata\n```\n\nTo install additional dependencies that will be used by the converters (e.g. `etl-manager` and `arrow` extras)\n\n```bash\npip install \'mojap-metadata[etl-manager,arrow] @ git+https://github.com/moj-analytical-services/mojap-metadata\'\n```\n\n<hr>\n\n# Metadata \n\nThis module creates a class called `Metadata` which allows you to interact with our agnostic metadata schemas. The `Metadata` class deals with parsing, manipulating and validating metadata json schemas.\n\n## The Schema\n\nOur metadata schemas are used to define a table. The idea of these schemas are to define the contexts of a table with generic metadata schemas. If you want to use this schema to interact with Oracle, PyArrow or AWS Glue for example, then you can create a Converter class to take the metadata and converter it to a schema that works with that tool (or vice versa).\n\nWhen adding a parameter to the metadata config first thing is to look if it exists in [json-schema](https://json-schema.org/understanding-json-schema/index.html). For example `enum`, `pattern` and `type` are parameters in our column types but come from json schema naming definitions.\n\nAn example of a basic metadata schema:\n\n```json\n{\n    "$schema" : "$schema": "https://moj-analytical-services.github.io/metadata_schema/mojap_metadata/v1.0.0.json",\n    "name": "employees",\n    "description": "table containing employee information",\n    "file_format": "parquet",\n    "columns": [\n        {\n            "name": "employee_id",\n            "type": "int64",\n            "type_desc": "integer",\n            "description": "an ID for each employee",\n            "minimum": 1000,\n            "maximum": 9999\n        },\n        {\n            "name": "employee_name",\n            "type": "string",\n            "type_string": "string",\n            "description": "name of the employee"\n        },\n        {\n            "name": "employee_dob",\n            "type": "date64",\n            "type_desc": "date",\n            "description": "date of birth for the employee in ISO format",\n            "pattern": "^\\\\d{4}-([0]\\\\d|1[0-2])-([0-2]\\\\d|3[01])$"\n        }\n    ]\n}\n```\n\n### Schema Properties\n\n- **name:** String that can be whatever you want to name the table. Best to avoid spaces as most systems do not like that but it will let you do this.\n\n- **file_format:** String denoting the file format.\n\n- **columns:** List of objects where each object descibes a column in your table. Each column object must have at least a `name` and a (`type` or `type_description`).\n\n    - **name:** String denoting the name of the column.\n    - **type:** String specifing the type the data is in. We use data types from the [Apache Arrow project](https://arrow.apache.org/docs/python/api/datatypes.html). We use their type names as it seems to comprehensively cover most of the data types we deal with. _Note: In our naming convention for types we allow `bool` (which is equivalent to `bool_`) and `list` (which is equivalent to `list_`)._\n    - **type_category:** These group different sets of `type` properties into a single superset. These are: `integer`, `float`, `string`, `timestamp`, `bool`, `list`, `struct`. For example we class `int8, int16, int32, int64, uint8, uint16, uint32, uint64` as `integer`. It allows users to give more generic types if your data is not coming from a system or output with strict types (i.e. data exported from Excel or an unknown origin). The Metadata class has default type values for each given `type_category`. See the `default_type_category_lookup` attribute of the `Metadata` class to see said defaults. This field is required if `type` is not set.\n    - **description:** Description of the column.\n    - **enum:** List of what values that column can take. _(Same as the standardised json schema keyword)._\n    - **pattern:** Regex pattern that value has to to match (for string type_categories only). _(Same as the standardised json schema keyword)._\n    - **minLength / maxLength:** The minimum and maximum length of the string (for string type_categories only). _(Same as the standardised json schema keyword)._\n    - **minimum / maximum:** The minumum and maximum value a numerical type can take (for integer and float type_categories only).\n- **partitions:** List of what columns in your dataset are partitions.\n- **table_location:** the location of the table. This is a string that can represent a file path, directory, url, etc.\n- **database_name:**  the name of the database this table belongs to.\n\n#### Additional Schema Parameters\n\nWe allow users to add addition parameters to the table schema object or any of the columns in the schema. If there are specific parameters / tags you want to add to your schema it should still pass validation (as long as the additional parameters are not the same name of ones already used in the schema).\n\n## Usage\n\n```python\nfrom mojap_metadata import Metadata\n\n# Generate basic Metadata Table from dict\nmeta1 = Metadata(name="test", columns=[{"name": "c1", "type": "int64"}, {"name": "c2", "type": "string"}])\n\nprint(meta1.name) # test\nprint(meta1.columns[0]) #\xa0{"name": "c1", "type": "int64"}\nprint(meta1.description) # ""\n\n# Generate meta from dict\nd = {\n    "name": "test",\n    "columns": [\n        {"name": "c1", "type": "int64"},\n        {"name": "c2", "type": "string"}\n    ]\n}\nmeta2 = Metadata.from_dict(d)\n\n# Read / write to json\nmeta3 = Metadata.from_json("path/to/metadata_schema.json")\nmeta3.name = "new_table"\nmeta3.to_json("path/to/new_metadata_schema.json")\n```\n\n## Added Class methods and properties\n\nThe metadata class has some methods and properties that are not part of the schema but helps organise and manage the schema.\n\n### Column Methods\n\nThe class has multiple methods to alter the columns list.\n\n- `column_names`: Get a list of column names\n- `update_column`: If column with name matches replace it otherwise add it to the end\n- `remove_column`: Remove column that matches the given name. Note if a name in the `partitions` property matches that name then it is also removed.\n\n```python\n    meta = Metadata(columns=[\n        {"name": "a", "type": "int8"},\n        {"name": "b", "type": "string"},\n        {"name": "c", "type": "date32"},\n    ])\n    meta.column_names # ["a", "b", "c"]\n \n    meta.update_column({"name": "a", "type": "int64"})\n    meta.columns[0]["type"] # "int64"\n\n    meta.update_column({"name": "d", "type": "string"})\n    meta.column_names # ["a", "b", "c", "d"]\n\n    meta.remove_column("d")\n    assert meta.column_names == ["a", "b", "c"]\n```\n\nThe metadata class is a subclass of MutableMappings, where keys are column names and values are column metadata. \n\n- A metadata column can be accessed using the column name as a key.\n- A new or existing column can be updated using the column name as a key. The key must match the column name. Calls update_column.\n- A column can be deleted. Calls remove_column.\n- Columns of metadata can be iterated over.\n- The length of metadata is defined as the number of columns.\n\n```python\n# Access a specific column\nmeta["c1"] # {"name": "c1", "type": "int64"}\n\n# Add a new column (key must match name)\nmeta["c3"] = {"name": "c3", "type": "bool"}\n\n# Delete a column\ndel meta["c3"]\n\n# Iterate over all columns\nfor col in meta:\n    print(f"column name:{col["name"]}, column type:{col["type"]}")\n\n# Get the number of columns\nlen(meta) # 3\n```\n\n### force_partition_order Property\n\nBy default this is set to None. However can be set to `"start"` or `"end"`. When set to None the Metadata Class will not track column order relative to partitions.\n\n> Note: For Athena we normally set partitions at the end.\n\n```python\nmeta = Metadata(columns=[\n        {"name": "a", "type": "int8"},\n        {"name": "b", "type": "string"},\n        {"name": "c", "type": "date32"},\n    ])\n\nmeta.partitions = ["b"]\nmeta.column_names # ["a", "b", "c"]\n```\n\nIf set to `"start"` or `"end"` then any changes to partitions will affect the column order.\n\n```python\nmeta.force_partition_order = "start"\nmeta.column_names # ["b", "a" ,"c"]\n```\n\n### Generating Metdata objects\n\n<hr>\n\n# Converters\n\nConverters takes a Metadata object and generates something else from it (or can convert something to a Metadata object). Most of the time your converter will convert our schema into another systems schema. \n\n# Usage\n\nFor example the `ArrowConverter` takes our schemas and converts them to a pyarrow schema:\n\n```python\nfrom mojap_metadata import Metadata\nfrom mojap_metadata.converters.arrow_converter import ArrowConverter\n\nd = {\n    "name": "test",\n    "columns": [\n        {"name": "c1", "type": "int64"},\n        {"name": "c2", "type": "string"},\n        {"name": "c3", "type": "struct<k1: string, k2:list<int64>>"}\n    ],\n    "file_format": "jsonl"\n}\nmeta = Metadata.from_dict(d)\n\nac = ArrowConverter()\narrow_schema = ac.generate_from_meta(meta)\n\nprint(arrow_schema) # Could use this schema to read in data as arrow dataframe and cast it to the correct types\n```\n\nAnother use for the arrow converter is to convert it back from an Arrow schema to our metadata. This is especially useful if you have nested data types that would be difficult to write out the full `STRUCT` / `LIST`. Instead you can let Arrow do that for you and then pass the agnostic metadata object into something like the Glue Converter to generate a schema for AWS Glue.\n\n```python\nimport pyarrow as pa\nimport pandas as pd\n\nfrom mojap_metadata.converters.arrow_converter import ArrowConverter\n\ndata = {\n    "a": [0,1],\n    "b": [\n        {"cat": {"meow": True}, "dog": ["bone", "fish"]},\n        {"cat": {"meow": True}, "dog": ["bone", "fish"]},\n    ]\n}\ndf = pd.DataFrame(data)\narrow_df = pa.Table.from_pandas(df)\nac = ArrowConverter()\nmeta = ac.generate_to_meta(arrow_df.schema)\n\nprint(meta.columns) # [{\'name\': \'a\', \'type\': \'int64\'}, {\'name\': \'b\', \'type\': \'struct<cat:struct<meow:bool>, dog:list<string>>\'}]\n```\n\nAnother example is the `GlueConverter` which takes our schemas and converts them to a dictionary that be passed to the glue_client to deploy a schema on AWS Glue.\n\n```python\nimport boto3\nfrom mojap_metadata import Metadata\nfrom mojap_metadata.converters.glue_converter import GlueConverter\n\nd = {\n    "name": "test",\n    "columns": [\n        {"name": "c1", "type": "int64"},\n        {"name": "c2", "type": "string"},\n        {"name": "c3", "type": "struct<k1: string, k2:list<int64>>"}\n    ],\n    "file_format": "jsonl"\n}\nmeta = Metadata.from_dict(d)\n\ngc = GlueConverter()\nboto_dict = gc.generate_from_meta(meta, )\nboto_dict = gc.generate_from_meta(meta, database_name="test_db", table_location="s3://bucket/test_db/test/")\n\nprint(boto_dict) \n\nglue_client = boto3.client("glue")\nglue_client.create_table(**boto_dict) # Would deploy glue schema based on our metadata\n```\n\nAll converter classes are sub classes of the `mojap_metadata.converters.BaseConverter`. This `BaseConverter` has no actual functionality but is a boilerplate class that ensures standardised attributes  for all added `Converters` these are:\n\n- **generate_from_meta:** (function) takes a Metadata object and returns whatever the converter is producing .\n\n- **generate_to_meta:** (function) takes Any object (normally another schema for another system or package) and returns our Metadata object. (i.e. the reverse of generate_from_meta).\n\n- **options:** (Data Class) that are the options for the converter. The base options have a `suppress_warnings` parameter but it doesn\'t mean call converters use this. To get a better understanding of setting options see the `GlueConverter` class or the `tests/test_glue_converter.py` to see how they are set.\n\nincluded alongside `GlueConverter` is `GlueTable` that can overlay a metadata object, dictionary, or path to metadata file. it has one method:\n- **generate_from_meta:** generates a glue table from the metadata object, dict, or string path, takes the following arguments:\n    - _metadata:_ the metadata object, dict, or string path that is to be overlaid\n    - _table\\_location:_ a kwarg, the location of the table data. This can also be a property of the metadata object, dict, or file\n    - _database\\_name:_ a kwarg, the name of the glue database to put the table. This can also be a property of the metadata object, dict, or file\n\n\n##\xa0Further Usage\n\nSee the [mojap-aws-tools repo](https://github.com/moj-analytical-services/mojap-aws-tools-demo) which utilises the converters a lot in different tutorials.\n\n## Contributing and Design Considerations\n\nEach new converter (if not expanding on existing converters) should be added as a new submodule within the parent `converters` module. This is especially true if the new converter has additional package dependencies. By design the standard install of this package is fairly lightweight. However if you needed the `ArrowConverter` you would need to install the additional package dependencies for the arrow converter:\n\n```bash\npip install \'mojap-metadata[arrow] @ git+https://github.com/moj-analytical-services/mojap-metadata\'\n```\n\nThis means we can continuely add converters (as submodules) and add optional package dependencies ([see pyproject.toml](./pyproject.toml) ) without making the default install any less lightweight. `mojap_metadata` would only error if someone tries to import a converter subclass that with having the additional dependencies dependencies installed).\n\n## Postgres Converter\n\nPostgres Converter provides the following functionality\n\n1. Conenction to postgres database\n2. Extract the metadata from the tables\n3. Convert the extracted ouptut into Metadata object \n\n- **get_object_meta** (function) takes the table name, schema name then the extracts the metadata from postgres database and \nconverts into Metadata object \n\n- **generate_to_meta:** (function) takes the database connection and returns a list of Metadata object for all the (non-system schemas) schemas and tables from the connection.\n',
    'author': 'MoJ Data Engineering',
    'author_email': 'dataengineering@digital.justice.gov.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
